/*
 * Copyright (C) 2009  Axel Kesseler
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 */
#include "stdafx.h"
#include "output.h"
#include <windows.h>
#include <stdio.h>
#include <process.h>
#include <assert.h>
#include <stdlib.h>

#include "messages.h"

using namespace plexdata::logging;

///////////////////////////////////////////////////////////////////////////
// Private singleton console wrapper class.
///////////////////////////////////////////////////////////////////////////
class CSingletonConsole
{
private:

	static const unsigned short MESSAGE_COLOR = FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_BLUE | FOREGROUND_INTENSITY;
	static const unsigned short WARNING_COLOR = FOREGROUND_GREEN | FOREGROUND_RED | FOREGROUND_INTENSITY;
	static const unsigned short ERROR_COLOR   = FOREGROUND_RED   | FOREGROUND_INTENSITY;
	static const unsigned short DUMP_COLOR    = FOREGROUND_GREEN | FOREGROUND_BLUE | FOREGROUND_INTENSITY;

	volatile long m_nRefCount;
	volatile bool m_bUseColor;

	unsigned long m_nOriginalMode;
	LPTSTR        m_pOriginalTitle;
	COORD         m_nOriginalSize;
	bool          m_bCanClose;

private:

	//
	// CSingletonConsole()
	//
	CSingletonConsole()
		: m_nRefCount(0)
		, m_nOriginalMode(0)
		, m_pOriginalTitle(NULL)
		, m_bCanClose(false)
		, m_bUseColor(true)
	{
		m_nOriginalSize.X = 0;
		m_nOriginalSize.Y = 0;
	};

public:

	//
	// ~CSingletonConsole()
	//
	virtual ~CSingletonConsole()
	{
		// Free memory used for the title.
		delete this->m_pOriginalTitle;
		this->m_pOriginalTitle = NULL;

		// Ensure the detach.
		this->ForceDetach();
	}

public:

	//
	// GetInstance()
	//
	static CSingletonConsole& GetInstance()
	{
		static CSingletonConsole theInstance;
		return theInstance;
	};

public:

	//
	// Attach()
	//
	bool Attach()
	{
		bool success = false;

		// Try to allocate a new console window.
		if ( ::AllocConsole() )
		{
			this->m_bCanClose = true; // Frees created console window during Detach().
			success = true;
		}
		else 
		{
			// A console window already exists.
			if ( ERROR_ACCESS_DENIED == ::GetLastError() )
			{
				success = true;
			}
		}

		if ( success )
		{
			if ( !::SetConsoleCtrlHandler( (PHANDLER_ROUTINE)CSingletonConsole::ConsoleCtrlHandler, TRUE ) )
			{
				success = false; // Attach() is unsuccessful if the handler registration failed!
			}
		}

		return success;
	}

	//
	// Message()
	//
	bool Message( LPCTSTR pOutput )
	{
		bool success = false;
		WORD nOldAttributes = 0;

		if ( this->m_nRefCount > 0 && pOutput != NULL )
		{
			HANDLE hConsole = ::GetStdHandle( STD_OUTPUT_HANDLE );
			size_t length   = _tcslen( pOutput );
			DWORD  written  = 0;

			// Ensure validity...
			assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );

			if ( this->m_bUseColor )
			{
				CONSOLE_SCREEN_BUFFER_INFO sInfo = {0};

				::GetConsoleScreenBufferInfo( hConsole, &sInfo );

				// Save original attributes for later restore.
				nOldAttributes = sInfo.wAttributes;

				// Remove all foreground information from original attributes.
				WORD nAttributes = nOldAttributes & ~( FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED );

				// Set new attributes using message depending color.
				::SetConsoleTextAttribute( hConsole, nAttributes | MESSAGE_COLOR );
			}

			// Write the output...
			if ( ::WriteConsole( hConsole, pOutput, (DWORD)length, &written, NULL ) )
			{
				success = true;
			}

			if ( this->m_bUseColor )
			{
				// Restore original attributes.
				::SetConsoleTextAttribute( hConsole, nOldAttributes );
			}
		}
		return success;
	}

	//
	// Warning()
	//
	bool Warning( LPCTSTR pOutput )
	{
		bool success = false;
		WORD nOldAttributes = 0;

		if ( this->m_nRefCount > 0 && pOutput != NULL )
		{
			HANDLE hConsole = ::GetStdHandle( STD_OUTPUT_HANDLE );
			size_t length   = _tcslen( pOutput );
			DWORD  written  = 0;

			// Ensure validity...
			assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );

			if ( this->m_bUseColor )
			{
				CONSOLE_SCREEN_BUFFER_INFO sInfo = {0};

				::GetConsoleScreenBufferInfo( hConsole, &sInfo );

				// Save original attributes for later restore.
				nOldAttributes = sInfo.wAttributes;

				// Remove all foreground information from original attributes.
				WORD nAttributes = nOldAttributes & ~( FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED );

				// Set new attributes using message depending color.
				::SetConsoleTextAttribute( hConsole, nAttributes | WARNING_COLOR );
			}

			// Write the output...
			if ( ::WriteConsole( hConsole, pOutput, (DWORD)length, &written, NULL ) )
			{
				success = true;
			}

			if ( this->m_bUseColor )
			{
				// Restore original attributes.
				::SetConsoleTextAttribute( hConsole, nOldAttributes );
			}
		}
		return success;
	}

	//
	// Error()
	//
	bool Error( LPCTSTR pOutput )
	{
		bool success = false;
		WORD nOldAttributes = 0;

		if ( this->m_nRefCount > 0 && pOutput != NULL )
		{
			HANDLE hConsole = ::GetStdHandle( STD_ERROR_HANDLE );
			size_t length   = _tcslen( pOutput );
			DWORD  written  = 0;

			// Ensure validity...
			assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );

			if ( this->m_bUseColor )
			{
				CONSOLE_SCREEN_BUFFER_INFO sInfo = {0};

				::GetConsoleScreenBufferInfo( hConsole, &sInfo );

				// Save original attributes for later restore.
				nOldAttributes = sInfo.wAttributes;

				// Remove all foreground information from original attributes.
				WORD nAttributes = nOldAttributes & ~( FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED );

				// Set new attributes using message depending color.
				::SetConsoleTextAttribute( hConsole, nAttributes | ERROR_COLOR );
			}

			// Write the output...
			if ( ::WriteConsole( hConsole, pOutput, (DWORD)length, &written, NULL ) )
			{
				success = true;
			}

			if ( this->m_bUseColor )
			{
				// Restore original attributes.
				::SetConsoleTextAttribute( hConsole, nOldAttributes );
			}
		}
		return success;
	}

	//
	// Dump()
	//
	bool Dump( LPCTSTR pOutput )
	{
		bool success = false;
		WORD nOldAttributes = 0;

		if ( this->m_nRefCount > 0 && pOutput != NULL )
		{
			HANDLE hConsole = ::GetStdHandle( STD_OUTPUT_HANDLE );
			size_t length   = _tcslen( pOutput );
			DWORD  written  = 0;

			// Ensure validity...
			assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );


			if ( this->m_bUseColor )
			{
				CONSOLE_SCREEN_BUFFER_INFO sInfo = {0};

				::GetConsoleScreenBufferInfo( hConsole, &sInfo );

				// Save original attributes for later restore.
				nOldAttributes = sInfo.wAttributes;

				// Remove all foreground information from original attributes.
				WORD nAttributes = nOldAttributes & ~( FOREGROUND_BLUE | FOREGROUND_GREEN | FOREGROUND_RED );

				// Set new attributes using message depending color.
				::SetConsoleTextAttribute( hConsole, nAttributes | DUMP_COLOR );
			}

			// Write the output...
			if ( ::WriteConsole( hConsole, pOutput, (DWORD)length, &written, NULL ) )
			{
				success = true;
			}

			if ( this->m_bUseColor )
			{
				// Restore original attributes.
				::SetConsoleTextAttribute( hConsole, nOldAttributes );
			}
		}
		return success;
	}

	//
	// Detach()
	//
	bool Detach()
	{
		if ( this->m_bCanClose && this->m_nRefCount <= 1 )
		{
			this->m_bCanClose = false;

			if ( ::FreeConsole() )
			{
				return true;
			}
			else 
			{
				return false; // Only fail in this case!
			}
		}
		else
		{
			return true; // Don't fail if already detached.
		}
	}

	//
	// Resize()
	//
	bool Resize( long& lines, long& columns )
	{
		bool success = false;
		CONSOLE_SCREEN_BUFFER_INFO sInfo = {0};

		HANDLE hConsole = ::GetStdHandle( STD_OUTPUT_HANDLE );

		// Ensure validity...
		assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );

		if ( this->m_nOriginalSize.X == 0 && this->m_nOriginalSize.Y == 0 )
		{
			// Get current screen buffer size (and other stuff).
			if ( ::GetConsoleScreenBufferInfo( hConsole, &sInfo ) )
			{
				this->m_nOriginalSize   = sInfo.dwSize;
			}
		}

		if ( ::GetConsoleScreenBufferInfo( hConsole, &sInfo ) )
		{
			// Get current size limits.
			int nWindowWidth  = ( sInfo.srWindow.Right - sInfo.srWindow.Left ) + 1;
			int nWindowHeight = ( sInfo.srWindow.Bottom - sInfo.srWindow.Top ) + 1;

			// Check requested size limits and adjust values if needed.
			if ( columns < nWindowWidth  ) { columns = nWindowWidth;  }
			if ( lines   < nWindowHeight ) { lines   = nWindowHeight; }

			// Change current screen buffer size independent of previous calls.
			sInfo.dwSize.X = static_cast<short>( columns );
			sInfo.dwSize.Y = static_cast<short>( lines );
			
			if ( ::SetConsoleScreenBufferSize( hConsole, sInfo.dwSize ) )
			{
				success = true;
			}
		}

		return success;
	}

	//
	// Colors()
	//
	void Colors( bool enable )
	{
		this->m_bUseColor = enable;
	}

	//
	// SetTitle()
	//
	void SetTitle( LPCTSTR title )
	{
		// Save original console title at first function call.
		if ( NULL == this->m_pOriginalTitle )
		{
			this->m_pOriginalTitle = new TCHAR[MAX_PATH+1];
			::GetConsoleTitle( this->m_pOriginalTitle, MAX_PATH );
		}

		::SetConsoleTitle( title );
	}

	//
	// RestoreTitle()
	//
	void RestoreTitle()
	{
		// Check for last attached console instance.
		if ( this->m_pOriginalTitle != NULL && this->m_nRefCount <= 1 )
		{
			// Restore original title.
			::SetConsoleTitle( this->m_pOriginalTitle );

			// Free unneeded memory.
			delete this->m_pOriginalTitle;
			this->m_pOriginalTitle = NULL;
		}
	}

	//
	// DisableDetach()
	//
	void DisableDetach()
	{
		if ( this->m_bCanClose )
		{
#if ( _WIN32_WINNT >= 0x0500 )
			HWND hWnd = ::GetConsoleWindow();
			if ( hWnd != NULL )
			{
				HMENU hMenu = ::GetSystemMenu( hWnd, false );
				if ( hMenu != NULL )
				{
					if ( ::DeleteMenu( hMenu, SC_CLOSE, MF_BYCOMMAND ) )
					{
						::DrawMenuBar( hWnd );
					}
				}
			}
#else
#pragma message ( "WARNING: Function ::AttachConsole() only available since Windows 2000 (_WIN32_WINNT >= 0x0500)" )
#endif // _WIN32_WINNT >= 0x0500
		}
	}

	//
	// EnableQuickEditMode()
	//
	void EnableQuickEditMode()
	{
		HANDLE hConsole = ::GetStdHandle( STD_INPUT_HANDLE );

		// Ensure validity...
		assert( hConsole != NULL && hConsole != INVALID_HANDLE_VALUE );

		if ( this->m_nOriginalMode == 0 )
		{
			::GetConsoleMode( hConsole, &m_nOriginalMode );
		}

		// Ensure quick edit mode independent of previous calls.
		::SetConsoleMode( hConsole, ENABLE_QUICK_EDIT_MODE | ENABLE_EXTENDED_FLAGS );
	}

	//
	// RestoreCharacteristics()
	//
	void RestoreCharacteristics()
	{
		if ( this->m_nOriginalMode != 0 )
		{
			// Restore original console mode.
			::SetConsoleMode( ::GetStdHandle( STD_INPUT_HANDLE ), m_nOriginalMode );
			this->m_nOriginalMode = 0;
		}

		if ( this->m_nOriginalSize.X != 0 && this->m_nOriginalSize.Y != 0 )
		{
			// Restore original screen buffer size.
			::SetConsoleScreenBufferSize( ::GetStdHandle( STD_OUTPUT_HANDLE ), this->m_nOriginalSize );	
			this->m_nOriginalSize.X = 0;
			this->m_nOriginalSize.Y = 0;
		}
	}

	//
	// AddRef()
	//
	void AddRef()
	{
		m_nRefCount = ::InterlockedIncrement( &m_nRefCount );
	}

	//
	// Release()
	//
	long Release()
	{
		if ( m_nRefCount > 0 )
		{
			m_nRefCount = ::InterlockedDecrement( &m_nRefCount );
		}
		return m_nRefCount;
	}

private:

	//
	// ForceDetach()
	//
	void ForceDetach()
	{
		this->m_nRefCount = 0; // Forces detach!
		this->Detach();
	}

	//
	// ConsoleCtrlHandler()
	//
	static BOOL ConsoleCtrlHandler( DWORD dwCtrlType )
	{
		switch( dwCtrlType ) 
		{ 
			case CTRL_C_EVENT: 
			case CTRL_BREAK_EVENT: 
			{
				if ( CSingletonConsole::GetInstance().m_bCanClose )
				{
					return TRUE; // Suppress CTRL+C/BREAK in case of a Windows application.
				}
				else
				{
					return FALSE; // Forward CTRL+C/BREAK in case of a console application.
				}
			}
			case CTRL_CLOSE_EVENT: // Doesn't happen if current process is a Windows application!
			case CTRL_LOGOFF_EVENT: 
			case CTRL_SHUTDOWN_EVENT: 
			default: 
			{
				return FALSE; 
			}
		}
	}
};

///////////////////////////////////////////////////////////////////////////
// 
///////////////////////////////////////////////////////////////////////////

COutput::COutput()
	: m_szName(NULL)
	, m_nLength(0)
	, m_nWidth(0)
	, m_bColors(false)
{
}

COutput::~COutput()
{
	delete this->m_szName;
}

bool COutput::Name( LPCTSTR name )
{
	// Delete memory of old name.
	delete this->m_szName;
	this->m_szName = NULL;

	// Recreate name value only if needed.
	if ( name != NULL )
	{
		size_t length = _tcslen( name );

		this->m_szName = new TCHAR[length+1];

		if ( this->m_szName != NULL ) // Safety check!
		{
			return ( 0 == _tcsncpy_s( this->m_szName, length+1, name, length ) );
		}
		else
		{
			return false; // Out of memory!
		}
	}
	else
	{
		return true;
	}
}

bool COutput::Extract( LPTSTR& pBuffer, long& count, LPCTSTR pFormat, va_list argList )
{
	bool   success = false;
	int    length  = 0;
	double dummy   = 0.0; // Load floating point library, otherwise runtime error R6002 will occur!

	if ( pFormat != NULL )
	{
		length = _vsctprintf( pFormat, argList ) + 1; // _vsctprintf doesn't include terminating '\0'.

		if ( length > 0 )
		{
			pBuffer = new TCHAR[length];

			if ( pBuffer != NULL )
			{
				::ZeroMemory( pBuffer, length );

				// Try to format given parameter list.
				count = _vstprintf_s( pBuffer, length, pFormat, argList ) + 1; // _vstprintf_s doesn't include terminating '\0'.

				// Set function call result.
				success = ( count > 0 );
			}
		}
	}

	return success;
}

bool COutput::PrepareDump( LPTSTR& pOutput, const void* pBuffer, unsigned long size, bool absolute )
{
	bool success = false;

	const int BYTES_PER_LINE = 16;

	const int ADDR_SIZE = ( sizeof(void*) * 2 + 1); // 8 chars per address + 1 space.
	const int BYTE_SIZE = ( sizeof(unsigned char) * 2 + 1 ) * BYTES_PER_LINE; // 2 chars per byte + 1 space.
	const int CHAR_SIZE = sizeof(unsigned char) * BYTES_PER_LINE; // 1 char per byte.
	const int LINE_SIZE = ADDR_SIZE + BYTE_SIZE + CHAR_SIZE + 2 + 1; // All pieces + 2 spaces + 1 new line.

	int lines = size / BYTES_PER_LINE + ( ( ( size % BYTES_PER_LINE ) != 0 ) ? 1 : 0 );
	int count = lines * LINE_SIZE + 1; // Memory for last new line is needed!

	if ( pBuffer != NULL && size > 0 )
	{
		pOutput = new TCHAR[count+1];

		if ( pOutput != NULL )
		{
			::ZeroMemory( pOutput, count+1 );

			unsigned char* pOffset   = reinterpret_cast<unsigned char*>( const_cast<void*>( pBuffer ) );
			unsigned char* pPosition = pOffset;

			for ( int line = 0; line < lines; line++ )
			{
				TCHAR szLine[LINE_SIZE+1] = _T("");
				TCHAR szAddr[ADDR_SIZE+1] = _T(""); // Address(hex) + Space
				TCHAR szByte[BYTE_SIZE+1] = _T(""); // 16 * ( Byte(hex) + Space )
				TCHAR szChar[CHAR_SIZE+1] = _T(""); // 16 * ( Byte(chr) )

				// Format address string for current line.
				_stprintf_s( szAddr, ADDR_SIZE+1, _T("%8.8X "), ( absolute ? pOffset : NULL ) + ( pPosition - pOffset ) );

				// Format byte string and character string for current line.
				for ( int index = 0; index < BYTES_PER_LINE; index++ )
				{
					if ( pPosition - pOffset < static_cast<long>(size) )
					{
						TCHAR szCurrent[5] = _T("");

						// Format and add current byte to current line.
						_stprintf_s( szCurrent, sizeof(szCurrent)/sizeof(TCHAR), _T("%2.2X "), *pPosition & 0x000000FF );

						_tcscat_s( szByte, BYTE_SIZE+1, szCurrent );

						// Format and add current character to current line.
						szCurrent[0] = _T('\0');

						if ( ( *pPosition & 0x000000FF ) < 0x20 )
						{
							_tcscat_s( szCurrent, sizeof(szCurrent)/sizeof(TCHAR), _T(".") );
						}
						else
						{
							_stprintf_s( szCurrent, sizeof(szCurrent)/sizeof(TCHAR), _T("%c"), *pPosition & 0x000000FF );
						}

						_tcscat_s( szChar, CHAR_SIZE+1, szCurrent );

						// Move to next byte.
						pPosition++;
					}
					else
					{
						_tcscat_s( szByte, BYTE_SIZE+1, _T("   ") );
						_tcscat_s( szChar, CHAR_SIZE+1, _T(" ") );
					}
				}

				// Put everything together.
				_stprintf_s( szLine, LINE_SIZE+1, _T("%s %s %s\n"), szAddr, szByte, szChar );

				// Add current line to output buffer.
				_tcscat_s( pOutput, count+1, szLine );
			}

			success = true;
		}
	}

	return success;
}

///////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////

CConsole::CConsole() 
	: COutput()
{
}

CConsole::~CConsole()
{
	this->Detach();
}

bool CConsole::Attach( LPCTSTR title, bool overwrite /*=false*/ )
{
	bool success = false;

	if ( CSingletonConsole::GetInstance().Attach() )
	{
		if ( this->Name( title ) )
		{
			CSingletonConsole::GetInstance().AddRef();
			CSingletonConsole::GetInstance().SetTitle( title );
			CSingletonConsole::GetInstance().EnableQuickEditMode();
			CSingletonConsole::GetInstance().DisableDetach();
			success = true;
		}
	}

	return success;
}

bool CConsole::Message( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput  = NULL;
	long    length   = 0;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		// Don't convert extracted string because of some unsolved problems with console and UNICODE.
		success = CSingletonConsole::GetInstance().Message( pBuffer );
	}

	delete pBuffer;

	va_end( argList );

	return success;
}

bool CConsole::Warning( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput  = NULL;
	long    length   = 0;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		// Don't convert extracted string because of some unsolved problems with console and UNICODE.
		success = CSingletonConsole::GetInstance().Warning( pBuffer );
	}

	delete pBuffer;

	va_end( argList );

	return success;
}

bool CConsole::Error( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput  = NULL;
	long    length   = 0;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		// Don't convert extracted string because of some unsolved problems with console and UNICODE.
		success = CSingletonConsole::GetInstance().Error( pBuffer );
	}

	delete pBuffer;

	va_end( argList );

	return success;
}

bool CConsole::Dump( const void* pBuffer, unsigned long size, bool absolute )
{
	bool success = false;
	LPTSTR pOutput = NULL;

	if ( this->PrepareDump( pOutput, pBuffer, size, absolute ) )
	{
		// Don't convert extracted string because of some unsolved problems with console and UNICODE.
		success = CSingletonConsole::GetInstance().Dump( pOutput );
	}

	delete pOutput;

	return success;
}

bool CConsole::Detach()
{
	bool success = false;

	if ( CSingletonConsole::GetInstance().Detach() )
	{
		CSingletonConsole::GetInstance().RestoreCharacteristics();
		CSingletonConsole::GetInstance().RestoreTitle();
		CSingletonConsole::GetInstance().Release();
		success = true;
	}

	return success;
}

bool CConsole::Resize( long lines, long columns )
{
	this->m_nLength = lines;
	this->m_nWidth  = columns;

	return CSingletonConsole::GetInstance().Resize( this->m_nLength, this->m_nWidth );
}

void CConsole::Colors( bool enable )
{
	this->m_bColors = enable;

	CSingletonConsole::GetInstance().Colors( this->m_bColors );
}

///////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////

CLogfile::CLogfile() 
	: COutput()
	, m_hFile(INVALID_HANDLE_VALUE)
{
}

CLogfile::~CLogfile()
{
	this->Detach();
}

bool CLogfile::Attach( LPCTSTR file, bool overwrite )
{
	bool success    = false;
	DWORD access    = GENERIC_READ|GENERIC_WRITE;
	DWORD share     = FILE_SHARE_READ|FILE_SHARE_WRITE;
	DWORD attribute = FILE_ATTRIBUTE_NORMAL;
	DWORD open      = overwrite ? CREATE_ALWAYS : OPEN_ALWAYS;

	if ( this->m_hFile == INVALID_HANDLE_VALUE )
	{
		if ( this->Name( file ) )
		{
			this->m_hFile = ::CreateFile( file, access, share, NULL, open, attribute, NULL );

			if ( this->m_hFile != INVALID_HANDLE_VALUE )
			{
				::SetFilePointer( this->m_hFile, 0, 0, FILE_END ); // Move to EOF.
				success = true;
			}
		}
	}

	return success;
}

bool CLogfile::Message( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput = NULL;
	size_t  length  = 0;
	DWORD   written = 0;
	va_list argList;

	if ( this->m_hFile != INVALID_HANDLE_VALUE )
	{
		va_start( argList, format );

		if ( this->Extract( pBuffer, count, format, argList ) )
		{
			this->Convert( pBuffer, count, pOutput, length );

			if ( ::WriteFile( this->m_hFile, pOutput, (DWORD)length, &written, NULL ) )
			{
				::FlushFileBuffers( this->m_hFile );
				success = true;
			}

			delete pOutput; // Don't delete pBuffer (See function Convert())!
		}

		va_end( argList );
	}
	
	return success;
}

bool CLogfile::Warning( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput = NULL;
	size_t  length  = 0;
	DWORD   written = 0;
	va_list argList;

	if ( this->m_hFile != INVALID_HANDLE_VALUE )
	{
		va_start( argList, format );

		if ( this->Extract( pBuffer, count, format, argList ) )
		{
			this->Convert( pBuffer, count, pOutput, length );

			if ( ::WriteFile( this->m_hFile, pOutput, (DWORD)length, &written, NULL ) )
			{
				::FlushFileBuffers( this->m_hFile );
				success = true;
			}

			delete pOutput; // Don't delete pBuffer (See function Convert())!
		}

		va_end( argList );
	}
	
	return success;
}

bool CLogfile::Error( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
    LPTSTR  pBuffer = NULL;
	void*   pOutput = NULL;
	size_t  length  = 0;
	DWORD   written = 0;
	va_list argList;

	if ( this->m_hFile != INVALID_HANDLE_VALUE )
	{
		va_start( argList, format );

		if ( this->Extract( pBuffer, count, format, argList ) )
		{
			this->Convert( pBuffer, count, pOutput, length );

			if ( ::WriteFile( this->m_hFile, pOutput, (DWORD)length, &written, NULL ) )
			{
				::FlushFileBuffers( this->m_hFile );
				success = true;
			}

			delete pOutput; // Don't delete pBuffer (See function Convert())!
		}

		va_end( argList );
	}
	
	return success;
}

bool CLogfile::Dump( const void* pBuffer, unsigned long size, bool absolute )
{
	bool   success = false;
	LPTSTR pDump = NULL;
	void*  pOutput = NULL;
	DWORD  written = 0;
	size_t length  = 0;

	if ( this->m_hFile != INVALID_HANDLE_VALUE )
	{
		if ( this->PrepareDump( pDump, pBuffer, size, absolute ) )
		{
			this->Convert( pDump, _tcslen(pDump), pOutput, length );

			if ( ::WriteFile( this->m_hFile, pOutput, (DWORD)length, &written, NULL ) )
			{
				::FlushFileBuffers( this->m_hFile );
				success = true;
			}

			delete pOutput; // Don't delete pDump (See function Convert())!
		}
	}

	return success;
}

bool CLogfile::Detach()
{
	bool success = false;

	if ( this->m_hFile != INVALID_HANDLE_VALUE )
	{
		if ( ::CloseHandle( this->m_hFile ) )
		{
			this->m_hFile = INVALID_HANDLE_VALUE;
			success = true;
		}
	}

	return success;
}

bool CLogfile::Resize( long nSize, long dontCare/*=0*/ )
{
	this->m_nLength = nSize;
	this->m_nWidth  = dontCare;

	// TODO: Implementation!
	::SetLastError( ERROR_CALL_NOT_IMPLEMENTED );

	return false;
}

void CLogfile::Colors( bool dontCare/*=false*/ )
{
	// REMARK: We don't need colors until a log file supports differen colors.
	this->m_bColors = dontCare;
}

void CLogfile::Convert( LPTSTR& pBuffer, size_t count, void*& pOutput, size_t& length )
{
	// Initialize result values.
	pOutput = reinterpret_cast<void*>(pBuffer);
	length  = ( count - 1 ) * sizeof(TCHAR); // Don't convert terminating '\0'

	if ( sizeof(TCHAR) == sizeof(wchar_t) ) // Fits if _UNICODE is defined!
	{
		LPSTR pConvert = new char[count];

		if ( pConvert != NULL ) // If pConvert is NULL the original buffer will be written.
		{
			::ZeroMemory( pConvert, count );

			::WideCharToMultiByte(
				CP_ACP,
				0,
				reinterpret_cast<LPCWSTR>(pBuffer),
				-1, // Assume, string ends with terminating '\0'
				pConvert,
				(int)count,
				NULL,
				NULL
			);

			delete pBuffer;	// Free original buffer created by function Extract()

			// Reassign converted buffer information to result values.
			pOutput = reinterpret_cast<void*>(pConvert);
			length  = ( count - 1 ) * sizeof(char); // Don't write terminating '\0'
		}
	}
}

///////////////////////////////////////////////////////////////////////////
//
///////////////////////////////////////////////////////////////////////////

const LPTSTR CEventlog::REGPATH = _T("SYSTEM\\CurrentControlSet\\Services\\EventLog\\Application\\%s");

CEventlog::CEventlog() 
	: COutput()
	, m_hSource(NULL)
	, m_resourceBinary(NULL)
{
}

CEventlog::~CEventlog()
{
	this->Detach();
	delete this->m_resourceBinary;
}

bool CEventlog::Attach( LPCTSTR source, bool overwrite /*=false*/ )
{
	bool success = false;

	if ( this->m_hSource == NULL )
	{
		if ( this->Name( source ) )
		{
			// Adminstrator rights AND the "rigth" module file name 
			// (binary which links msg00000.bin etc.) are necessary!
			this->RegisterEventSourceRegistry();

			this->m_hSource = ::RegisterEventSource( NULL, this->Name() );

			if ( this->m_hSource != NULL )
			{
				success = true;
			}
		}
	}

	return success;
}

bool CEventlog::Message( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
	TCHAR*  pBuffer = NULL;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		if ( this->m_hSource != NULL )
		{
			LPCTSTR ppMessages[1] = {pBuffer};

			// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/reporting_an_event.htm
			BOOL bReported = ::ReportEvent(
				this->m_hSource,           // Event log handle. 
				EVENTLOG_SUCCESS,          // Event type. 
				NULL,                      // Event category.  
				EVENTLOG_MESSAGE,          // Event identifier. 
				NULL,                      // No user security identifier. 
				1,                         // Number of substitution strings. 
				0,                         // No data. 
				ppMessages,                // Pointer to message strings. 
				NULL                       // No pointer to raw data
			);

			success = ( bReported == TRUE ) ? true : false;
		}

		delete pBuffer;
	}

	va_end( argList );

	return success;
}

bool CEventlog::Warning( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
	TCHAR*  pBuffer = NULL;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		if ( this->m_hSource != NULL )
		{
			LPCTSTR ppMessages[1] = {pBuffer};

			// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/reporting_an_event.htm
			BOOL bReported = ::ReportEvent(
				this->m_hSource,           // Event log handle. 
				EVENTLOG_WARNING_TYPE,     // Event type. 
				NULL,                      // Event category.  
				EVENTLOG_WARNING,          // Event identifier. 
				NULL,                      // No user security identifier. 
				1,                         // Number of substitution strings. 
				0,                         // No data. 
				ppMessages,                // Pointer to message strings. 
				NULL                       // No pointer to raw data
			);

			success = ( bReported == TRUE ) ? true : false;
		}

		delete pBuffer;
	}

	va_end( argList );

	return success;
}

bool CEventlog::Error( LPCTSTR format, ... )
{
	bool    success = false;
	long    count   = 0;
	TCHAR*  pBuffer = NULL;
	va_list argList;

	va_start( argList, format );

	if ( this->Extract( pBuffer, count, format, argList ) )
	{
		if ( this->m_hSource != NULL )
		{
			LPCTSTR ppMessages[1] = {pBuffer};

			// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/reporting_an_event.htm
			BOOL bReported = ::ReportEvent(
				this->m_hSource,           // Event log handle. 
				EVENTLOG_ERROR_TYPE,       // Event type. 
				NULL,                      // Event category.  
				EVENTLOG_ERROR,            // Event identifier. 
				NULL,                      // No user security identifier. 
				1,                         // Number of substitution strings. 
				0,                         // No data. 
				ppMessages,                // Pointer to message strings. 
				NULL                       // No pointer to raw data
			);

			success = ( bReported == TRUE ) ? true : false;
		}

		delete pBuffer;
	}

	va_end( argList );

	return success;
}

bool CEventlog::Dump( const void* pBuffer, unsigned long size, bool absolute )
{
	bool success = false;

	if ( pBuffer != NULL && size > 0 )
	{
		if ( this->m_hSource != NULL )
		{
			TCHAR szMessage[50] = {_T('\0')};

#pragma warning( disable : 4996 )
			// REMARK: Function _stprintf_s() manipulates the given pointer in an unacceptable way!
			//         => This means: Don't use function _stprintf_s() right here!
			_stprintf( szMessage, _T("Address: 0x%8.8X Size: %d"), pBuffer, size );
#pragma warning( default : 4996 )

			LPCTSTR ppMessages[1] = { szMessage };

			// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/reporting_an_event.htm
			BOOL bReported = ::ReportEvent(
				this->m_hSource,           // Event log handle. 
				EVENTLOG_INFORMATION_TYPE, // Event type. 
				NULL,                      // Event category.  
				EVENTLOG_DUMP,             // Event identifier. 
				NULL,                      // No user security identifier. 
				1,                         // Number of substitution strings. 
				size,                      // Size of raw data buffer data. 
				ppMessages,                // Pointer to message strings. 
				const_cast<void*>(pBuffer) // Pointer to raw data
			);

			success = ( bReported == TRUE ) ? true : false;
		}
	}

	return success;
}

bool CEventlog::Detach()
{
	bool success = false;

	if ( this->m_hSource != NULL )
	{
		if ( ::DeregisterEventSource( this->m_hSource ) )
		{
			this->m_hSource = NULL;
			success = true;
		}

		this->UnregisterEventSourceRegistry();
	}

	return success;
}


bool CEventlog::Resize( long lengthDonCare/*=0*/, long widthDontCare/*=0*/ )
{
	this->m_nLength = lengthDonCare;
	this->m_nWidth  = widthDontCare;

	return true;
}

void CEventlog::Colors( bool dontCare/*=false*/ )
{
	// REMARK: Support of "different" colors is already handled by the message types.
	this->m_bColors = dontCare;
}

bool CEventlog::MessageResourceBinary( HMODULE hResourceBinary )
{
	bool   success = false;
	DWORD  nSize   = MAX_PATH;
	LPTSTR pFile   = new TCHAR[nSize+1];

	while ( !success )
	{
		::GetModuleFileName( hResourceBinary, pFile, nSize );

		if ( ::GetLastError() == ERROR_INSUFFICIENT_BUFFER )
		{
			delete pFile;

			nSize *= 2L;
			pFile = new TCHAR[nSize+1];

		}
		else if ( ::GetLastError() == ERROR_SUCCESS )
		{
			success = this->MessageResourceBinary( pFile );
			delete pFile;
		}
		else
		{
			break; // End loop in any other case.
		}
	}
	
	return success;
}

bool CEventlog::MessageResourceBinary( LPCTSTR resourceBinary )
{
	bool success = false;

	delete this->m_resourceBinary;
	this->m_resourceBinary = NULL;

	// Recreate module file value only if needed.
	if ( resourceBinary != NULL )
	{
		size_t length = _tcslen( resourceBinary );

		this->m_resourceBinary = new TCHAR[length+1];

		if ( this->m_resourceBinary != NULL ) // Safety check!
		{
			if ( 0 == _tcsncpy_s( this->m_resourceBinary, length+1, resourceBinary, length ) )
			{
				success = true;
			}
		}
	}

	return success;
}

bool CEventlog::RegisterEventSourceRegistry()
{
	//
	// REMARK: How does Windows event source work?
	//         The manipulation of the Windows Registry is not really needed 
	//         to get a Windows event source working! Furthermore, Windows event 
	//         source needs some regisrty information (see below) to display 
	//         event data in an appropriated form. This means a Windows event
	//         source works in every case independently of the existance of 
	//         those registry information. So we don't really need to manipulate 
	//         the Windows registry...
	//

#ifdef _USE_REGISTRY_EVENT_DATA_ 

	// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/adding_a_source_to_the_registry.htm

	TCHAR  resourceBinary[MAX_PATH+1] = _T("");
	TCHAR  regSubKey[MAX_PATH+1]      = _T(""); 
	HKEY   hKey            = NULL; 
	LPTSTR pResourceBinary = NULL;
	DWORD  nResourceBinary = 0L;
	DWORD  nRegData        = 0L; 
	LONG   result          = 0L;

	// Prepare needed module file name.
	if ( this->m_resourceBinary != NULL )
	{
		pResourceBinary = this->m_resourceBinary;
		nResourceBinary = ( _tcslen( pResourceBinary ) + 1 ) * sizeof(TCHAR);
	}
	else
	{
		// TODO: Validate function result.
		if ( this->MessageResourceBinary( ::GetModuleHandle( NULL ) ) )
		{
			pResourceBinary = this->m_resourceBinary;
			nResourceBinary = ( _tcslen( pResourceBinary ) + 1 ) * sizeof(TCHAR);
		}
		else
		{
			return false;
		}
	}

	// Create needed event source registry subkey. 
	_stprintf_s( regSubKey, MAX_PATH, CEventlog::REGPATH, this->Name() ); 

	// Create message source's base registry key.
	result = ::RegCreateKeyEx( 
		HKEY_LOCAL_MACHINE, 
		regSubKey, 
		0, 
		NULL, 
		REG_OPTION_NON_VOLATILE, 
		KEY_WRITE, 
		NULL, 
		&hKey, 
		&nRegData
	);

	if ( result != ERROR_SUCCESS )
	{
		return false;
	}

	// Set message file name.
	result = ::RegSetValueEx(
		hKey,                     // subkey handle 
		_T("EventMessageFile"),   // value name 
		0,                        // must be zero 
		REG_EXPAND_SZ,            // value type 
		(LPBYTE)pResourceBinary,  // data pointer
		(DWORD)nResourceBinary    // data size
	);

	if ( result != ERROR_SUCCESS )
	{
		::RegCloseKey( hKey ); 
		return false;
	}

	// Set supported event types. 
	nRegData = EVENTLOG_ERROR_TYPE | EVENTLOG_WARNING_TYPE | EVENTLOG_INFORMATION_TYPE; 
	result = ::RegSetValueEx(
		hKey,                     // subkey handle 
		_T("TypesSupported"),     // value name 
		0,                        // must be zero 
		REG_DWORD,                // value type 
		(LPBYTE)&nRegData,        // data pointer
		sizeof(DWORD)             // data size
	);

	if ( result != ERROR_SUCCESS )
	{
		::RegCloseKey( hKey ); 
		return false;
	}

	// Set category message file.
	result = ::RegSetValueEx(
		hKey,                     // subkey handle 
		_T("CategoryMessageFile"),// value name 
		0,                        // must be zero 
		REG_EXPAND_SZ,            // value type 
		(LPBYTE)pResourceBinary,  // data pointer
		(DWORD)nResourceBinary    // data size
	);

	if ( result != ERROR_SUCCESS )
	{
		::RegCloseKey( hKey ); 
		return false;
	}

	// Set number of categories.
	nRegData = 1L;
	result = ::RegSetValueEx(
		hKey,                      // subkey handle 
		_T("CategoryCount"),       // value name 
		0,                         // must be zero 
		REG_DWORD,                 // value type 
		(LPBYTE)&nRegData,         // data pointer
		sizeof(DWORD)              // data size
	);

	if ( result != ERROR_SUCCESS )
	{
		::RegCloseKey( hKey ); 
		return false;
	}

	::RegCloseKey( hKey ); 

	return true;

#else 

	return true;

#endif // _USE_REGISTRY_EVENT_DATA_
}

bool CEventlog::UnregisterEventSourceRegistry()
{
	//
	// REMARK: How does Windows event source work?
	//         See comments in function RegisterEventSourceRegistry()...
	//

#ifdef _USE_REGISTRY_EVENT_DATA_ 

	// See: ms-help://MS.VSCC.v90/MS.MSDNQTR.v90.en/eventlog/base/adding_a_source_to_the_registry.htm

	bool  success = false;
	TCHAR regSubKey[MAX_PATH+1] = _T(""); 

	// Create needed event source registry subkey. 
	_stprintf_s( regSubKey, MAX_PATH, CEventlog::REGPATH, this->Name() ); 

	// Remove event source from registry. 
	if ( ERROR_SUCCESS == ::RegDeleteKey( HKEY_LOCAL_MACHINE, regSubKey ) )
	{
		success = true;
	}

	return success;

#else 

	return true;

#endif // _USE_REGISTRY_EVENT_DATA_
}

